
class Icon {
	string name;
	string nameLower;
	int code;
	string icon;
	string hex;
	string escape;
	
	Icon(string name, int code) {
		this.name = name;
		this.nameLower = name.ToLower();
		this.code = code;
		
		// convert code point to UTF-8
		if(0x0000 <= code && code <= 0x007f) {
			this.icon = " ";
			this.icon[0] = 0b00000000 | (0b01111111 & (code >>  0));
		} else if(0x00000080 <= code && code <= 0x000007ff) {
			this.icon = "  ";
			this.icon[0] = 0b11000000 | (0b00011111 & (code >>  6));
			this.icon[1] = 0b10000000 | (0b00111111 & (code >>  0));
		} else if(0x00000800 <= code && code <= 0x0000ffff) {
			this.icon = "   ";
			this.icon[0] = 0b11100000 | (0b00001111 & (code >> 12));
			this.icon[1] = 0b10000000 | (0b00111111 & (code >>  6));
			this.icon[2] = 0b10000000 | (0b00111111 & (code >>  0));
		} else if(0x00010000 <= code && code <= 0x0010ffff) {
			this.icon = "    ";
			this.icon[0] = 0b11110000 | (0b00000111 & (code >> 18));
			this.icon[1] = 0b10000000 | (0b00111111 & (code >> 12));
			this.icon[2] = 0b10000000 | (0b00111111 & (code >>  6));
			this.icon[3] = 0b10000000 | (0b00111111 & (code >>  0));
		} else {
			throw("IconList: Icon out of range: " + name + " = " + code);
		}
		
		hex = "";
		for(int i = 0; i < icon.Length; i++) {
			if(i > 0) hex += " ";
			hex += Text::Format("%02x", this.icon[i]);
		}
		
		escape = "\\u" + Text::Format("%04x", this.code);
	}
	
	bool matches(string search) {
		return this.nameLower.Contains(search) || this.hex.Contains(search) || this.escape.Contains(search);
	}
}

enum ESortColumn {
	File,
	Name,
	Code,
	UTF8
}

bool windowOpen = false;
bool helpOpen = false;
array<Icon@> icons = {};
array<Icon@> iconsSort;
array<Icon@> iconsFilter;
ESortColumn sortCol = ESortColumn(-1);
string searchStr = "";

int _parseHexDigit(uint8 digit) {
	if(digit < 0x40) return digit & 0b1111;
	else return (digit & 0b1111) + 9;
}

int parseHex(string hex) {
	if(!Regex::IsMatch(hex, "(0x)?[0-9a-fA-F]{1,8}")) {
		error("invalid 32-bit hex! \"" + hex + "\"");
		return 0;
	}
	int index = 0;
	int value = 0;
	if(hex.StartsWith("0x")) {
		index = 2;
	}
	for(; index < hex.Length; index++) {
		value <<= 4;
		value += _parseHexDigit(hex[index]);
	}
	return value;
}

void RenderMenu() {
	if (UI::MenuItem("\\$ddd" + Icons::Info + "\\$z List All Icons", "", windowOpen)) {
		if(!windowOpen) {
			auto app = cast<CTrackMania>(GetApp());
			
			icons.Resize(0);
			IO::FileSource iconsFile("res\\Icons.txt");
			string prefix = "";
			while(!iconsFile.EOF()) {
				string line = iconsFile.ReadLine();
				
				if(line.Contains("DEPRECATED")) {
					break; // all icons below the deprecations message should not be used
					
				} else if(line.Contains("namespace")) {
					prefix += line.Trim().Split(" ")[1] + "::";
					
				} else if(line.Contains("string")) {
					array<string> chunks = line.Trim().Split(" ");
					string name = chunks[1];
					int code = parseHex("0x" + chunks[3].SubStr(3, 4));
					if(code != 0) {
						icons.InsertLast(Icon(prefix + name, code));
					}
					
				} else if(line.Contains("}")) {
					int index = prefix.Length - 2; // skip terminal "::"
					while(--index >= 0 && prefix[index] != ":"[0]);
					prefix = prefix.SubStr(0, index+1);
				}
			}
			iconsSort = icons;
			iconsFilter = iconsSort;
			
			windowOpen = true;
		} else {
			searchStr = "";
			
			windowOpen = false;
		}
	}
}

void RenderInterface() {
	if(windowOpen) {
		bool needFilter = false;
		
		UI::SetNextWindowSize(600, 700);
		UI::SetNextWindowPos(200, 200, UI::Cond::Once);
		UI::Begin("\\$ddd" + Icons::Info + "\\$z Icon Listing", windowOpen);
		
		string newSearchStr = UI::InputText("Filter", searchStr, UI::InputTextFlags(UI::InputTextFlags::AutoSelectAll | UI::InputTextFlags::NoUndoRedo));
		
		if(UI::Selectable("\\$777Click an entry to copy to clipboard. Click here for instructions.\\$z", false)) {
			helpOpen = !helpOpen;
		}
		UI::Separator();
		
		if(UI::BeginTable("tableau", 4, UI::TableFlags(UI::TableFlags::Sortable | UI::TableFlags::NoSavedSettings | UI::TableFlags::BordersInnerV | UI::TableFlags::SizingStretchProp | UI::TableFlags::ScrollY))) {
			UI::TableSetupScrollFreeze(0, 1);
			UI::TableSetupColumn("Icon", UI::TableColumnFlags::DefaultSort, 15.f, ESortColumn::File);
			UI::TableSetupColumn("Identifier", UI::TableColumnFlags::None, 45.f, ESortColumn::Name);
			UI::TableSetupColumn("Code point", UI::TableColumnFlags::None, 20.f, ESortColumn::Code);
			UI::TableSetupColumn("UTF-8 bytes", UI::TableColumnFlags::None, 20.f, ESortColumn::UTF8);
			UI::TableHeadersRow();
			
			UI::TableSortSpecs@ sortSpecs = UI::TableGetSortSpecs();
			if(sortSpecs !is null && sortSpecs.Specs.Length == 1) {
				if(sortSpecs.Dirty) {
					if(sortCol == ESortColumn(sortSpecs.Specs[0].ColumnUserID)) {
						iconsSort.Reverse();
					} else {
						sortCol = ESortColumn(sortSpecs.Specs[0].ColumnUserID);
						switch(sortCol) {
							case ESortColumn::File:
								iconsSort = icons;
								break;
							case ESortColumn::Name:
								iconsSort.Sort(function(a,b) { return a.name < b.name; });
								break;
							case ESortColumn::Code:
							case ESortColumn::UTF8:
								iconsSort.Sort(function(a,b) { return a.code < b.code; });
								break;
						}
					}
					
					needFilter = true;
					sortSpecs.Dirty = false;
				}
			}
			
			if(needFilter || newSearchStr != searchStr) {
				searchStr = newSearchStr;
				string searchStrLower = searchStr.ToLower();
				iconsFilter = {};
				for(uint i = 0; i < iconsSort.Length; i++) {
					if(searchStrLower == "" || iconsSort[i].matches(searchStrLower)) {
						iconsFilter.InsertLast(iconsSort[i]);
					}
				}
			}
			
			
			UI::ListClipper clipper(iconsFilter.Length);
			while(clipper.Step()) {
				for(int i = clipper.DisplayStart; i < clipper.DisplayEnd; i++) {
					UI::TableNextRow();
					UI::TableNextColumn();
					copyableItem(iconsFilter[i].icon);
					UI::TableNextColumn();
					copyableItem(iconsFilter[i].name);
					UI::TableNextColumn();
					copyableItem(iconsFilter[i].escape);
					UI::TableNextColumn();
					copyableItem(iconsFilter[i].hex);
				}
			}
			
			UI::EndTable();
		}
		
		UI::End();
	}
	
	if(helpOpen) {
		vec2 mouse = UI::GetMousePos();
		UI::SetNextWindowPos(int(mouse.x), int(mouse.y), UI::Cond::Appearing);
		UI::Begin("\\$ddd" + Icons::Info + "\\$z Icon Listing Help", helpOpen, UI::WindowFlags(UI::WindowFlags::NoCollapse | UI::WindowFlags::AlwaysAutoResize | UI::WindowFlags::NoSavedSettings));
		
		UI::Markdown("To use these icons, simply reference the listed Identifier in your script. Since Openplanet version 1.22.0, it is no longer necessary to import Icons.as as a dependency.\n\n" +
		             "See [https://openplanet.dev/docs/reference/icons](https://openplanet.dev/docs/reference/icons) for additional information.\n\n" +
		             "These icons are intended for use in Openplanet plugins, and may be recolored like normal text.\n\n" + 
		             "The fonts required to render these icons are provided by Openplanet. Therefore, they may not display correctly for some players if used for other purposes, such as map names or in chat, and should not be used for such.\n\n" +
		             "For game-supported symbols, select from the list shown at [https://openplanet.nl/symbols](https://openplanet.nl/symbols).");
		UI::End();
	}
}

void copyableItem(const string &in content) {
	if(UI::Selectable(highlightMatch(content, searchStr), false)) {
		IO::SetClipboard(content);
		UI::ShowNotification("Copied!");
	}
}

string highlightMatch(const string &in input, const string &in search) {
	if(search.Length == 0) {
		return input;
	} else {
		return Regex::Replace(input, "(" + search + ")", "\\$f00$1\\$z", Regex::Flags(Regex::Flags::CaseInsensitive | Regex::Flags::ECMAScript));
	}
}
